import base64
import datetime
import glob
import httplib
import itertools
import json
import logging
import os
import random
import re
import socket
import stat
import sys
import time

import dns.resolver
import netifaces
import pam
import psutil
import rq
import tornado.escape
import tornado.web
from flask import Flask, request, send_file, session, make_response
from redis import Redis
from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
import tornado.platform.epoll

import _globals
import db
import http_client
import jobs
import settings
import trust
import util
import validators
from  ConfigParser import NoOptionError
from config_wrapper import config, CONFIG_FILES, SYSTEM_CONF_INI
from decorators import (authenticate, render_as_json, restart_service,
                        manager_route, appliance_only_route)
from errors import Error
from _globals import ADMIN_USER, DEFAULT_PASSWORD, make_timestamp
from safly.config import get_bind_address, get_build_id
from safly.container.os_helpers import get_monotonic_time
from safly.event_reporter import EventReporter
from safly.saml_auth import (SAMLAuth, InvalidSamlResponse, InvalidSamlConfig,
                             get_cert_and_key)
from safly.secret_util import secret_util
from safly.tenant_config import (get_normalized_attributes, get_saml_config,
                                 TenantConfigIni,
                                 validate_attribute_authorization)
from safly import cert_utils, net_utils, upgrade
from _token import Token

# Used because the saml_auth module expects a request object
# from Tornado, but the dashboard is a flask application.
# TODO: Remove this class once the dashboard has been migrated
# to be a Tornado application
class TornadoRequest(object):
   def __init__(self, req):
      self.host = req.host
      self.path = req.path
      self.body_arguments = {key: [req.form[key]] for key in req.form}
      self.query_arguments = {key: [req.args[key]] for key in req.args}

class SAMLAuthError(Exception):
   def __init__(self, value):
      self.value = value
      Exception.__init__(self, value)

# TODO: change cookie name to better match functionality. Needs to happen in
# parallel with several PnR repositories.
PNR_ADMIN_AUTH_COOKIE = 'pnr_id'
PNR_ADMIN_INFO_COOKIE = 'pnr.userId'
PNR_ADMIN_SAML_ERROR_COOKIE = 'saml.error'
PNR_ADMIN_TEMP_COOKIE = 'pnr.temp'

# Error message strings
ERR_PORT_RESERVED = 'cannot use reserved port number: %s'
ERR_PORT_IN_USE = 'port: %s is in use by another process'
ERR_SAME_PORT = 'port numbers for %s and %s must be unique'

log_format = 'time="%(asctime)-15s" logger=%(name)s module=%(module)s level=%(levelname)-7s %(message)s'
logging.basicConfig(level=logging.INFO,
   format=log_format,
   filename=None,
   stream=sys.stderr,
   filemode='w')
logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)

app = Flask(__name__, static_folder='static/app', static_url_path='')

# FIXME SECURITY: This secret key needs to either be auto-generated each time
# the process starts or should move into a configuration file
app.config.update(
   DEBUG=config.getboolean('dashboard', 'debug'),
   SESSION_COOKIE_SECURE=True,
   SECRET_KEY=secret_util.get_secret('dashboard', 'secret')
)

app.reporter = EventReporter.instance()

# TODO: should also get the host and port from config.
app.redis = Redis(password=config.get('redis', 'password'),
                  socket_timeout=config.getfloat('redis', 'socket_timeout'),
                  socket_connect_timeout=config.getfloat(
                        'redis', 'socket_connect_timeout'),
                  socket_keepalive=True,
                  retry_on_timeout=True)
rq.use_connection(app.redis)


def _get_user_name():
   # FIXME: using username in the session cookie for easy access. The "right
   # way" would be to get the username directly from the token (see login).
   return session.get('username')

def _get_job_status_result(job_id, queue=jobs.DASHBOARD_Q):
   job = rq.Queue(queue).fetch_job(job_id)
   if job:
      val = { 'id': job.id, 'status': job.get_status() }
      if job.is_finished:
         val.update(job.result)
      return val
   return (Error.JOB_NOT_FOUND, 404)

def delete_failed_job(job, *exc_info):
   """
   This is custom exception handler to delete the failed job in RQ instead of
   storing it in the failed queue (default behavior).
   """
   logging.info('event="dashboard-job-failed" '
                'info="deleting failed job" job=%s', job.id)
   # prepares job for eventual deletion
   job.cleanup(ttl=15)
   # returns the flag as False to avoid the exception to fall through the
   # default exception handler (which stores such jobs in failed queue)
   return False

@app.after_request
def add_header(response):
   response.cache_control.no_cache = True
   response.cache_control.no_store = True
   response.cache_control.must_revalidate = True
   response.cache_control.max_age = 0
   return response

@app.route('/api/v1/ping')
@render_as_json
def api_v1_ping():
   return { "ok": True }

def _gen_login_info(username):
   token = Token.make(app, username)
   session['token'] = token
   # FIXME: storing username in the session cookie for easy access. The "right
   # way" would be to get the username directly from the token.
   session['username'] = username
   rval = {'token': token}
   # We initialize first_login here, instead of globally to make writing
   # nosetests easier
   if util.is_first_login():
      rval['chpass'] = True

   return rval

def _attach_auth_cookie_to_response(resp, username):
   # Create secure tornado cookie for PnR auth
   secure_cookie = tornado.web.create_signed_value(
         secret_util.get_secret('dashboard', 'secret'),
         PNR_ADMIN_AUTH_COOKIE,
         json.dumps({ 'v': 1, 'userId': username, 'tid': -1}))
   expiration = datetime.datetime.now() + datetime.timedelta(days=float(config.get('dashboard', 'cookie_expiration')))
   resp.set_cookie(PNR_ADMIN_AUTH_COOKIE,
         secure_cookie,
         secure=True,
         httponly=True,
         expires=expiration)
   insights_license_enabled = config.getboolean(util.DASHBOARD_SECTION, util.INSIGHTS_LICENSE_ENABLE)
   insights_retention_days = False
   if insights_license_enabled:
      druid_logging = util.get_logging_configuration().get('druid_logging', {})
      # Provide insights default 30 days retention options if druid_logging is not enabled
      insights_retention_days = druid_logging.get('max_days', 30) if druid_logging.get('enable') else 30

   # Create cookie to communicate to PnR frontend the user name and
   # capabilities.
   pnr_admin_info = tornado.escape.url_escape(
         json.dumps({
               'uid': username,
               'roles': json.loads(util.get_rbac_role(username)) if util.is_cmr() else None,
               # TODO: finalize capabilities, get them from INI (config)
               # any change in capabilities should also be updated in
               # safeview/service/src/dashboard/static/app/management-station/management-station.js
               'capabilities': {
                  'policy': True,
                  'userException': True,
                  'safemail': True,
                  'safemailSmtpLog': True,
                  'safemailSmtpDashboard': True,
                  'safemail_mobile': True,
                  'sslBumpBypass': True,
                  'categoryIsolate': True,
                  'safemailPolicy': True,
                  'bannerCustomization': True,
                  'dialogsCustomization': True,
                  'safedocs': True,
                  'av_scan': config.getboolean(util.SAFEFILE_SECTION, util.AVENGINE_LICENSE_ENABLE),
                  'sandbox': config.getboolean(util.SAFEFILE_SECTION, util.SANDBOX_LICENSE_ENABLE),
                  'fireeye': True, # FireEye feature available on-prem only
                  'generic_push': config.getboolean(util.SAFEFILE_SECTION, util.GENERIC_PUSH_LICENSE_ENABLE),
                  'filedownload': True,
                  'docviewIsolation': False,
                  'flash': True,
                  'webmail': False,
                  'fileupload': False,
                  'webReadonly': True,
                  'browserNonBrowser': True,
                  'bnbStrictMode': config.getboolean(util.DASHBOARD_SECTION, util.BNB_STRICT_MODE_UI_LICENSE_ENABLE),
                  'directOnAllowUI': config.getboolean(util.DASHBOARD_SECTION, util.DIRECT_ON_ALLOW_UI_LICENSE_ENABLE),
                  'skipCategorizationUI': config.getboolean(util.DASHBOARD_SECTION, util.SKIP_CATEGORIZATION_UI_LICENSE_ENABLE),
                  'mobileSupport': config.getboolean(util.DASHBOARD_SECTION, util.MOBILE_SUPPORT_LICENSE_ENABLE),
                  'ddl': True,
                  'rbac': True,
                  'reversingLabsLicense': config.getboolean(util.SAFEMAIL_SECTION, util.REVERSING_LABS_LICENSE_ENABLE),
                  'emailAttachments': config.getboolean(util.SAFEMAIL_SECTION, util.SAFEMAIL_ATTACHMENTS_LICENSE_ENABLE),
                  'emailAttachmentsSafePdf': True,
                  'secondaryManager' : util.is_secondary_cmr(),
                  'urlObfuscationException': True,
                  'applicationLinks': True,
                  'insights': insights_license_enabled,
                  'insightsRetentionDays': insights_retention_days,
               }}),
         plus=False)
   resp.set_cookie(PNR_ADMIN_INFO_COOKIE, pnr_admin_info, secure=True)
   # send login audit log event to policy
   if util.is_cmr():
      util.send_audit_log_to_policy('login', username)

   mode = config.get(util.DASHBOARD_SECTION, util.MODE_KEY)
   enabled_frontend = config.getboolean(util.DASHBOARD_SECTION, 'enable_pnr_frontend')
   if mode == 'management_station' or enabled_frontend:
      reporting_xsrf_response = jobs.get_reporting_xsrf()
      if reporting_xsrf_response.get('ok'):
         resp.set_cookie('_xsrf', reporting_xsrf_response.get('xsrf_cookie'), secure=True, expires=expiration)

   return resp


@app.route('/api/v1/saml')
@render_as_json
def api_v1_saml_get():
   conf_reader = util.ConfigReader('authentication')
   if not conf_reader.getboolean(util.USE_SAML_FOR_ADMIN_KEY):
      logging.warning('SAML Auth not enabled for admin')
      return (Error.ADMIN_SAML_DISABLED, 401)

   if not util._is_admin_saml_configured():
      logging.warning('Unable to find SAML configuration')
      return (Error.LOGIN_FAILED, 401)

   tornado_request = TornadoRequest(request)
   try:
      tenant_config = TenantConfigIni()
      admin_saml_conf = get_saml_config(tenant_config.tid_config[-1], 'admin')
      sp_cert, sp_key, sig_alg = (None, None, None)
      try:
         sp_cert, sp_key, sig_alg = get_cert_and_key(admin_saml_conf)
      except Exception:
         # cert or key could not be read, but we will still attempt a
         # request without signing it.
         logging.warning('Could not read SAML SP certificates.')
      saml_auth = SAMLAuth(tornado_request, request.host, admin_saml_conf,
                           {'https': True, 'server_port': 8443}, 'api/v1',
                           sp_cert=sp_cert, sp_key=sp_key, sig_alg=sig_alg)
   except (KeyError, NoOptionError, InvalidSamlConfig) as e:
      reason = getattr(e, 'error', None) or e.message
      logging.warning('Could not process SAML request: %s' % reason)
      raise SAMLAuthError('Could not process SAML request')
   relay_url = request.values.get('url', '/')
   login_url = saml_auth.get_login_url(relay_url)
   rval = {'redirect': login_url}
   resp = make_response(json.dumps(rval), 200, {'Content-Type': 'application/json'})
   return resp

@app.route('/api/v1/saml_check')
@render_as_json
def api_v1_saml_check_get():
   conf_reader = util.ConfigReader('authentication')
   if not conf_reader.getboolean(util.USE_SAML_FOR_ADMIN_KEY):
      saml_enabled = False
   elif not util._is_admin_saml_configured():
      saml_enabled = False
   else:
      saml_enabled = True

   resp = make_response(json.dumps(saml_enabled), 200, {'Content-Type': 'application/json'})
   return resp

@app.route('/api/v1/saml', methods=['POST'])
def api_v1_saml_post():
   resp = make_response(json.dumps({'redirect': True}), 302)
   resp.headers['location'] = 'https://' + request.host + '/d/#saml_login'
   conf_reader = util.ConfigReader('authentication')
   event_type = 'admin_auth_succeeded'
   reporting_data = {}
   reporting_data['details'] = 'Successful authentication via SAML.'
   reporting_data['client_ip'] = request.headers.get('X-Forwarded-For',
                                                     request.remote_addr)

   try:
      if not conf_reader.getboolean(util.USE_SAML_FOR_ADMIN_KEY):
         logging.warning('SAML Auth not enabled for admin')
         raise SAMLAuthError('SAML Auth not enabled for admin')

      if not util._is_admin_saml_configured():
         logging.warning('Unable to find SAML configuration')
         raise SAMLAuthError('Unable to find SAML configuration')

      tornado_request = TornadoRequest(request)
      trace = []
      try:
         tenant_config = TenantConfigIni()
         admin_saml_conf = get_saml_config(tenant_config.tid_config[-1], 'admin')
         sp_cert, sp_key, sig_alg = (None, None, None)
         try:
            sp_cert, sp_key, sig_alg = get_cert_and_key(admin_saml_conf)
         except Exception:
            # cert or key could not be read, but we will still attempt a
            # response check without the signing
            logging.warning('Could not read SAML SP certificates.')
         saml_auth = SAMLAuth(tornado_request, request.host, admin_saml_conf,
                              {'https': True, 'server_port': 8443}, 'api/v1',
                              sp_cert=sp_cert, sp_key=sp_key, sig_alg=sig_alg)
         saml_auth.process_response()
      except (KeyError, NoOptionError, InvalidSamlConfig, InvalidSamlResponse) as e:
         reason = getattr(e, 'error', None) or e.message
         logging.warning('Could not process SAML response: %s' % reason)
         raise SAMLAuthError('Could not process SAML response')

      uid = saml_auth.get_userid()
      reporting_data['user_id'] = uid

      logging.info('will be looking for %s = %s' %
                   (admin_saml_conf.get('authorized_attribute_name'),
                    admin_saml_conf.get('authorized_attribute_value')))
      saml_user = saml_auth.get_userid()
      attrs = get_normalized_attributes(saml_auth, admin_saml_conf, trace.append)

      is_authorized = validate_attribute_authorization(admin_saml_conf, 'admin',
                                                       attrs, trace.append)

      if util.is_cmr() and tenant_config.should_push_groups(
            tenant_config, admin_saml_conf, trace.append):
         if util.is_primary_cmr():
            # This has no return value to collect, and is not expected to throw.
            util.push_group_capabilities_sync(saml_user, -1, 'admin', attrs)
         if util.is_secondary_cmr():
            # secondary CMR needs to push groups to primary for rbac enforcement
            http_client.RedisMigrator(config,
                                      trust.MY_CLIENT_KEY,
                                      trust.MY_CLIENT_CERT,
                                      trust.TRUSTED_MGMT_SERVERS).migrate_admin_attributes_sync(saml_user, -1, 'admin', attrs)
      if is_authorized:
         resp = _attach_auth_cookie_to_response(resp, uid)
         frontend_info = _gen_login_info(uid)
         frontend_info['uid'] = uid
         frontend_cookie = tornado.escape.url_escape(json.dumps(frontend_info),
                                                     plus=False)
         resp.set_cookie(PNR_ADMIN_TEMP_COOKIE, frontend_cookie)

      else:
         logging.warning('User is not an authorized admin: %s' %
                         '; '.join(trace))
         raise SAMLAuthError('User is not an authorized admin')

   except SAMLAuthError as e:
      event_type = 'admin_auth_failed'
      reporting_data['details'] = 'Authentication via SAML failed'
      error_message = {'error': e.value}
      error_cookie = tornado.escape.url_escape(json.dumps(error_message),
                                               plus=False)
      resp.set_cookie(PNR_ADMIN_SAML_ERROR_COOKIE, error_cookie)

   app.reporter.report(event_type, reporting_data)
   return resp

@app.route('/api/v1/login', methods=["POST"])
@render_as_json
def api_v1_login():
   data = request.get_json()
   username = data.get('username', None)
   password = data.get('password', None)
   success = False
   conf_reader = util.ConfigReader('authentication')

   logging.info('Login attempt for %s', username)

   if not username or not password:
      logging.warning('login: missing username or password')
   elif username == ADMIN_USER:
      success = pam.authenticate(username, password)
   elif (conf_reader.getboolean('use_ldap_for_admin') and
         conf_reader.get('ldap_admin_group_name')):
      success = util.ldap_admin_login(username, password)
   else:
      logging.warning('login: unsupported username %s', username)

   event_type = 'admin_auth_failed'
   reporting_data = {}
   reporting_data['user_id'] = username
   reporting_data['client_ip'] = request.headers.get('X-Forwarded-For',
                                                     request.remote_addr)
   if success:
      event_type = 'admin_auth_succeeded'
      reporting_data['details'] = 'Successful authentication.'
   else:
      reporting_data['details'] = 'Authentication failed: Invalid username or password'
   app.reporter.report(event_type, reporting_data)

   if not success:
      return (Error.LOGIN_FAILED, 401)

   rval = _gen_login_info(username)
   # create a flask response object so we can set PnR cookies
   resp = make_response(json.dumps(rval),
                        200,
                        {'Content-Type': 'application/json'})
   # Create secure tornado cookie for PnR auth
   resp = _attach_auth_cookie_to_response(resp, username)
   return resp

@app.route('/api/v1/logout')
@render_as_json
@authenticate()
def api_v1_logout():
   session.pop('token', None)
   # Remove PnR admin cookies
   resp = make_response(json.dumps({'ok': True}), 200, { 'Content-Type': 'application/json' })
   resp.set_cookie(PNR_ADMIN_AUTH_COOKIE, '', expires=0)
   resp.set_cookie(PNR_ADMIN_INFO_COOKIE, '', expires=0)
   resp.set_cookie('_xsrf', '', expires=0)
   # send audit log logout event
   if util.is_cmr():
      util.send_audit_log_to_policy('logout', _get_user_name())
   return resp

@app.route('/api/v1/dashboard_timeout', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_dashboard_timeout():
   return {'timeout' : float(config.get('dashboard', 'cookie_expiration'))}

@app.route('/api/v1/cookie_refresh')
@render_as_json
@authenticate()
def api_v1_cookie_refresh():
   # Refresh expiration for PnR admin cookies
   resp = make_response(json.dumps({'ok': True}), 200, { 'Content-Type': 'application/json' })
   expiration = datetime.datetime.now() + datetime.timedelta(days=float(config.get('dashboard', 'cookie_expiration')))
   secure_cookie = tornado.web.create_signed_value(
         secret_util.get_secret('dashboard', 'secret'),
         PNR_ADMIN_AUTH_COOKIE,
         json.dumps({ 'v': 1, 'userId': _get_user_name(), 'tid': -1}))
   resp.set_cookie(PNR_ADMIN_AUTH_COOKIE,
         secure_cookie,
         secure=True,
         httponly=True,
         expires=expiration)

   userIdCookie = request.cookies.get(PNR_ADMIN_INFO_COOKIE)
   if userIdCookie:
      resp.set_cookie(PNR_ADMIN_INFO_COOKIE,
      userIdCookie,
      secure=True,
      expires=expiration)

   xsrf = request.cookies.get('_xsrf')
   if xsrf:
      old_xsrf = xsrf.split('|')
      old_xsrf[3] = str(int(time.time()))
      new_xsrf = '|'.join(old_xsrf)
      resp.set_cookie('_xsrf', new_xsrf, secure=True, expires=expiration)

   token = Token.make(app, _get_user_name())
   session['token'] = token
   return resp

@app.route('/api/v1/about')
@render_as_json
@authenticate()
def api_v1_about():
   return util.get_product_info()

@app.route('/api/v1/status')
@render_as_json
@authenticate()
def api_v1_status():
   return util.get_status()

@app.route('/api/v1/jobs/<job_id>/status')
@render_as_json
@authenticate()
def api_v1_get_job_status(job_id):
   """ Generic function for getting job status and result """
   return _get_job_status_result(job_id)

@app.route('/api/v1/license')
@render_as_json
@authenticate()
def api_v1_license_get():
   # since the license is set by a background worker, re-read config files to
   # ensure that the value is current.
   config.read(CONFIG_FILES)
   return {'license_string': config.get(util.DASHBOARD_SECTION,
                                        util.LICENSE_STRING),
           util.REVERSING_LABS: config.getboolean(util.SAFEMAIL_SECTION, util.REVERSING_LABS_LICENSE_ENABLE),
           util.AV_UPDATES: config.getboolean(util.SAFEFILE_SECTION, util.AVENGINE_LICENSE_ENABLE),
           util.AV_SANDBOX: config.getboolean(util.SAFEFILE_SECTION, util.SANDBOX_LICENSE_ENABLE),
           util.GENERIC_PUSH: config.getboolean(util.SAFEFILE_SECTION, util.GENERIC_PUSH_LICENSE_ENABLE),
           util.ATTACHMENTS: config.getboolean(util.SAFEMAIL_SECTION, util.SAFEMAIL_ATTACHMENTS_LICENSE_ENABLE),
           util.MIC_THREATDATA: config.getboolean(util.SAFEMAIL_SECTION, util.MIC_THREATDATA_LICENSE_ENABLE),
           util.CHAINED_CONNECT: config.getboolean(util.PNR_SQUID_SECTION, util.CHAINED_CONNECT_LICENSE_ENABLE),
           util.DIRECT_ON_ALLOW: config.getboolean(util.DASHBOARD_SECTION, util.DIRECT_ON_ALLOW_UI_LICENSE_ENABLE),
           util.BNB_STRICT_MODE: config.getboolean(util.DASHBOARD_SECTION, util.BNB_STRICT_MODE_UI_LICENSE_ENABLE),
           util.SKIP_CATEGORIZATION: config.getboolean(util.DASHBOARD_SECTION, util.SKIP_CATEGORIZATION_UI_LICENSE_ENABLE),
           util.MOBILE_SUPPORT: config.getboolean(util.DASHBOARD_SECTION, util.MOBILE_SUPPORT_LICENSE_ENABLE),
           util.INSIGHTS: config.getboolean(util.DASHBOARD_SECTION, util.INSIGHTS_LICENSE_ENABLE),
           'licensed_features_active': config.get('dashboard','licensed_features_active'),
           'licensed_features_expected': config.get('dashboard','licensed_features_expected')}

@app.route('/api/v1/license', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_license_put():
   job = jobs.check_save_license.delay(validators.get_string('license_string'))
   return {'id': job.id}

@app.route('/api/v1/validate_license')
@render_as_json
@authenticate()
def api_v1_validate_license_get():
   # since the license is set by a background worker, re-read config files to
   # ensure that the value is current.
   config.read(CONFIG_FILES)
   l = config.get(util.DASHBOARD_SECTION, util.LICENSE_STRING)
   if not l:
      return (Error.MISSING_LICENSE, 400)
   job = jobs.check_save_license.delay(l, save=False)
   return {'id': job.id}

@app.route('/api/v1/validate_license', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_validate_license_put():
   # Validate but do not save license
   job = jobs.check_save_license.delay(validators.get_string('license_string'), save=False)
   return {'id': job.id}

@app.route('/api/v1/settings/network')
@render_as_json
@authenticate()
def api_v1_get_settings_network():
   rval = {}
   rval.update(util.get_upstream_proxy_configuration())
   valid_ifaces = ['eth0', 'eth1']
   interfaces = [val for val in netifaces.interfaces() if val in valid_ifaces]
   inbound = config.get('service', 'inbound_iface')
   for interface in interfaces:
      conf_details = {}
      rval[interface] = conf_details
      static = net_utils.is_static(interface)
      connected = net_utils.is_interface_connected(interface)
      iface_conf = netifaces.ifaddresses(interface)
      # get the mac address associated with the interface
      mac = iface_conf[netifaces.AF_LINK][0]
      # get the ip family associated with the interface
      # note: netifaces.AF_INET is not always present
      #       for non-static ifaces
      ip_family = iface_conf.get(netifaces.AF_INET)
      if ip_family:
         conf_details.update({
            'ip': ip_family[0]['addr']
         })
      host = config.get('hostnames', interface)
      if (not host and interface == inbound):
         host = socket.gethostname()
      conf_details.update({
         'mac': mac['addr'],
         'is_connected': connected,
         'hostname': host
      })
      if interface != inbound:
         enabled = net_utils.is_interface_enabled(interface)
         conf_details['enable'] = enabled
         if not enabled:
            continue
      if static:
         # ip_family will exist for static ifaces
         net_conf = ip_family[0]
         conf_details.update({
            'dhcp': False,
            'netmask': net_conf['netmask']
         })
         if interface == inbound:
            gw = netifaces.gateways()['default'][netifaces.AF_INET][0]
            # By default, Resolver will use /etc/resolv.conf.
            ns = dns.resolver.Resolver().nameservers
            # Iterate over nameservers in case somehow an invalid,
            # comma-separated list of nameservers has somehow made its way
            # to the actual resolver nameserver settings.  This should never
            # happen, but extra caution here won't hurt, either, until all
            # DNS related issues are fixed.
            ns = [ip for ip in re.findall(r'[^,\s]+', ','.join(ns))]
            conf_details.update({
               'gateway': gw,
               'nameserver': ns[0] if len(ns) > 0 else '',
               'nameserver_secondary': ', '.join(ns[1:])
            })
      else:
         conf_details['dhcp'] = True
   return rval

@app.route('/api/v1/settings/network', methods=["PUT"])
@render_as_json
@authenticate()
@restart_service
def api_v1_put_settings_network():
   data = request.get_json()
   (upstream_proxy, http_host, http_port, upstream_proxy_username,
      upstream_proxy_password) = validators.get_upstream_proxy(
         'upstream_proxy', data=data, api_validation=True,
         enable_key='upstream_proxy', host_key='http_host',
         port_key='http_port', user_key='upstream_proxy_username',
         pass_key='upstream_proxy_password')
   allow_https_to_upstream_proxy = validators.get_bool(
      'allow_https_to_upstream_proxy', required=False) or False
   if upstream_proxy:
      jobs.configure_proxy.delay(http_host, http_port,
                                 upstream_proxy_username,
                                 upstream_proxy_password)
      jobs.restart_proxy_clients.delay()
   util.set_upstream_proxy_configuration(http_host or '', http_port or '',
                                         upstream_proxy,
                                         upstream_proxy_username,
                                         upstream_proxy_password,
                                         allow_https_to_upstream_proxy)
   interfaces = {}
   hosts = []
   inbound = config.get('service', 'inbound_iface')
   for key, conf in data.items():
      # checks for the interfaces in request body
      if not isinstance(conf, dict):
         continue
      interfaces[key] = {}
      enable = validators.get_bool('enable', data=conf, required=False)
      interfaces[key]['enable'] = enable
      if enable is False:
         continue
      host_new = validators.get_local_hostname('hostname', data=conf)
      host_old = net_utils.get_hostname() if (
                 key == inbound) else config.get('hostnames', key)
      if host_new != host_old:
         if key == inbound:
            net_utils.update_hostname(host_new)
         else:
            util.set_hostname_sys_config('hostnames', key, host_new)
         if key == config.get('service', 'management_iface'):
            hosts.append(host_new)
      dhcp = validators.get_bool('dhcp', data=conf)
      interfaces[key]['dhcp'] = dhcp
      if not dhcp:
         ip = validators.get_ip('ip', data=conf)
         netmask = validators.get_netmask('netmask', data=conf)
         interfaces[key].update({'ip': ip, 'netmask': netmask})
         if 'gateway' in conf:
            gateway = validators.get_ip('gateway', data=conf)
            ns = validators.get_ip('nameserver', data=conf)
            ns2 = validators.get_ip_list('nameserver_secondary',
                                         data=conf, required=False)
            if ns2:
               ns = ','.join([ns] + ns2)
            interfaces[key].update({'ns': ns, 'gateway': gateway})
   if hosts:
      trust.gen_dashboard_server_key_cert(overwrite=True, restart_haproxy=True,
                                          hosts=hosts)
   jobs.configure_network.delay(interfaces)
   # Even though these settings are not stored in the INI file, "touch" it so
   # configuration modified time is updated.
   os.utime(SYSTEM_CONF_INI, None)
   return { 'ok': True }

@app.route('/api/v1/settings/service_domain')
@render_as_json
@authenticate()
def api_v1_get_settings_service():
   return {'service_domain': config.get(util.SERVICE_SECTION, 'service_domain'),
           'login_host': config.get(util.SERVICE_SECTION, 'login_host'),
           'integrity_host_enabled': config.getboolean(util.AUTHENTICATION_SECTION, 'integrity_host_enabled'),
           'integrity_host': config.get(util.AUTHENTICATION_SECTION, 'integrity_host'),
           'use_integrity_token': config.getboolean(util.AUTHENTICATION_SECTION, 'use_integrity_token')}

# FIXME: this probably should be a POST.
@app.route('/api/v1/settings/service_domain', methods=['PUT'])
@render_as_json
@authenticate()
@restart_service
def api_v1_put_settings_service():
   # A "service domain" is a FQDN, use the get_host validator.
   service_domain = validators.get_host('service_domain')
   login_host = validators.get_host('login_host')
   integrity_host_enabled = validators.get_bool('integrity_host_enabled')
   integrity_host = validators.get_host('integrity_host')
   use_integrity_token = validators.get_bool('use_integrity_token')

   # TODO: Create new haproxy cert for service_domain
   # If domain is changing, build a new haproxy certificate
   # service_domain is new service domain
   old_service_domain = config.get('service', 'service_domain')
   old_login_host = config.get('service', 'login_host')
   # Generate haproxy certificate for new domain
   if service_domain != old_service_domain or login_host != old_login_host:
      logging.info('event=front-door-cert-update old-service-domain=%s '
                   'old-login-host=%s service-domain=%s login-host=%s',
                   old_service_domain, old_login_host, service_domain,
                   login_host)
      jobs.generate_haproxy_cert.delay()

   util.set_service_settings(service_domain, login_host,
                             integrity_host_enabled, integrity_host, use_integrity_token)
   # Changing the "service_domain" requires everything to be restarted.  As we
   # add to the "service settings" we might consider selectively restarting
   # based on what has changed.
   jobs.upstart_restart_all.delay()
   return {'ok': True}

@app.route('/api/v1/settings/logging')
@render_as_json
@authenticate()
def api_v1_get_settings_logging():
   return util.get_logging_configuration()

@app.route('/api/v1/settings/logging', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_logging():
   data = request.get_json()
   druid_logging = data.get('druid_logging')
   if druid_logging:
      validators.get_bool('enable', data=druid_logging)
      validators.get_int('max_days', min=1, max=180, data=druid_logging)
   all_entries = []
   _session = data.get('session')
   if _session:
      if isinstance(_session, dict):
         _session = [_session]
      all_entries.append(_session)
   system = data.get('system')
   if system:
      if isinstance(system, dict):
         system = [system]
      all_entries.append(system)
   opslog_enabled = data.get('opslog_enabled')
   operational = data.get('operational')
   if operational:
      if isinstance(operational, dict):
         operational = [operational]
      all_entries.append(operational)

   # Chain iterates each element in turn
   for entry in itertools.chain(*all_entries):
      validators.get_host('host', data=entry)
      validators.get_enum('protocol', data=entry, values=['tcp', 'udp'])
      validators.get_enum('format', data=entry, values=['CEF', 'KVP', 'LEEF'])
      validators.get_port('port', data=entry)

   metrics = data.get('metrics')
   #TODO: Better Validation for empty string and null values
   if metrics:
      validators.get_optional_host('graphite_host', data=metrics, required=False)
      validators.get_port('graphite_port', data=metrics, required=False)
      validators.get_optional_string('graphite_prefix', data=metrics, required=False)
      validators.get_optional_string('graphite_protocol',
                                     data=metrics, values=['tcp', 'udp'], required=False)
   snmp_v2 = data.get('snmp_v2')
   if snmp_v2:
      validators.get_host('host', data=snmp_v2)
      validators.get_port('port', data=snmp_v2)
      validators.get_string('community', data=snmp_v2)
   util.set_logging_configuration(druid_logging=druid_logging, session=_session,
                                  system=system, snmp_v2=snmp_v2, operational=operational,
                                  metrics=metrics, opslog_enabled=opslog_enabled)
   jobs.restart_logging_processes.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/time')
@render_as_json
@authenticate()
def api_v1_get_settings_time():
   ntp = net_utils.get_ntp()
   time_settings = util.get_time_configuration()
   time_settings['ntp'] = ntp
   time_settings['current_unix_time'] = time.time()
   return time_settings

@app.route('/api/v1/settings/time', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_time():
   ntp = validators.get_host('ntp', required=False)
   net_utils.update_ntp(ntp)

   timezone = validators.get_timezone('timezone')
   time_format = validators.get_enum('time_format', values=['hh', 'HH'])

   util.set_time_configuration(timezone, time_format)
   jobs.restart_logging_processes.delay()
   jobs.restart_fileserver_process.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/smtp')
@render_as_json
@authenticate()
def api_v1_get_settings_smtp():
   smtp_settings = util.get_smtp_configuration()
   return smtp_settings

@app.route('/api/v1/settings/smtp', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_smtp():
   server_address = validators.get_host('server_address')
   port = validators.get_port('port')
   username = validators.get_string('username')
   password = validators.get_string('password')
   from_address = validators.get_string('from_address')

   util.set_smtp_configuration(server_address, port, username, password, from_address)
   jobs.upstart_restart_pnr.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/pnr_proxy')
@render_as_json
@authenticate()
def api_v1_get_settings_pnr_proxy():
   return util.get_pnr_proxy_configuration()

@app.route('/api/v1/settings/pnr_proxy', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_pnr_proxy():
   error = {'reason': ''}
   no_bump_enabled = validators.get_bool('no_bump_enabled')
   no_bump_port = validators.get_port('no_bump_port')
   ssl_bump_enabled = validators.get_bool('ssl_bump_enabled')
   ssl_bump_port = validators.get_port('ssl_bump_port')
   listen_proxy_protocol = validators.get_bool('listen_proxy_protocol')
   proxy_protocol_trusted_ips = validators.get_optional_ip_list_string('proxy_protocol_trusted_ips')
   if ssl_bump_port == no_bump_port:
      error['reason'] = ERR_SAME_PORT % ('proxy port', 'ssl inspection port')
      return (error, 400)
   if util.is_reserved_port(no_bump_port):
      error['reason'] = ERR_PORT_RESERVED % no_bump_port
      return (error, 400)
   if util.is_reserved_port(ssl_bump_port):
      error['reason'] = ERR_PORT_RESERVED % ssl_bump_port
      return (error, 400)
   if util.is_port_in_use(no_bump_port, util.PNR_SQUID_PROCESS):
      error['reason'] = ERR_PORT_IN_USE % no_bump_port
      return(error, 400)
   if util.is_port_in_use(ssl_bump_port, util.PNR_SQUID_PROCESS):
      error['reason'] = ERR_PORT_IN_USE % ssl_bump_port
      return(error, 400)
   util.set_pnr_proxy_configuration(no_bump_enabled, no_bump_port,
                                    ssl_bump_enabled, ssl_bump_port,
                                    listen_proxy_protocol, proxy_protocol_trusted_ips)
   jobs.restart_squid_pnr_proxy.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/web')
@render_as_json
@authenticate()
def api_v1_get_settings_web():
   return util.get_web_configuration()

@app.route('/api/v1/settings/web', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_web():
   kwargs = dict()
   wc = 'weak_capture'
   aic = 'allow_invalid_certificates'
   rc = 'referrer_check'
   utl = 'tab_limit'
   dp = 'detach_profile'
   afd = 'allow_file_downloads'
   eu = 'enable_uploads'
   rolm = 'readonly_learning_mode'
   sff = 'send_forwarded_for'
   sxau = 'send_x_authenticated_user'
   svcs = 'ssl_validate_cert_san'
   kwargs[wc] = validators.get_bool(wc)
   kwargs[aic] = validators.get_enum(aic, values=[True, False, 'ask'])
   kwargs[rc] = validators.get_bool(rc)
   kwargs[utl] = validators.get_int(utl, min=1, max=1000)
   kwargs[dp] = validators.get_enum(dp, values=['disable', 'moderate'])
   kwargs[rolm] = validators.get_bool(rolm)
   kwargs[afd] = validators.get_bool(afd)
   kwargs[eu] = validators.get_bool(eu)
   kwargs[sff] = validators.get_bool(sff)
   kwargs[sxau] = validators.get_bool(sxau)
   kwargs[svcs] = validators.get_bool(svcs)
   util.set_web_configuration(**kwargs)
   jobs.restart_surrogates.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/auth/ldap')
@render_as_json
@authenticate()
def api_v1_get_settings_auth_ldap():
   return util.get_ldap_configuration()

def _get_ldap_parameters():
   kwargs = {}
   kwargs['ldap_host'] = validators.get_host('ldap_host')
   kwargs['ldap_port'] = validators.get_port('ldap_port')
   kwargs['ldap_mode'] = validators.get_enum(
      'ldap_mode', values=['ldap', 'ldaps', 'ldap_starttls'])

   kwargs['ldap_domain_dict'] = validators.get_json_dict('ldap_domain_dict')
   kwargs['ldap_default_domain'] = validators.get_string('ldap_default_domain')
   kwargs['ldap_use_custom_filter_clause'] = validators.get_bool(
      'ldap_use_custom_filter_clause')
   if kwargs['ldap_use_custom_filter_clause']:
      kwargs['ldap_custom_search_attribute'] = validators.get_string('ldap_custom_search_attribute')
   return kwargs

@app.route('/api/v1/settings/auth/ldap', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_auth_ldap():
   kwargs = _get_ldap_parameters()
   util.set_ldap_configuration(**kwargs)
   jobs.restart_auth_server.delay()
   return { 'ok': True }

@app.route('/api/v1/test/ldap', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_test_ldap():
   try:
      host = validators.get_host('ldap_host')
      port = validators.get_port('ldap_port')
      mode = validators.get_enum('ldap_mode',
                                 values=['ldap', 'ldaps', 'ldap_starttls'])
      username = validators.get_string('username')
      password = validators.get_string('password')
      val = util.test_ldap_configuration(mode, host, port, username, password)
      return {'domains': val}
   except util.TestError as e:
      return ({'error': 'test_failed', 'reason': e.message}, 400)

@app.route('/api/v1/settings/auth/saml')
@render_as_json
@authenticate()
def api_v1_get_settings_auth_saml():
   saml_settings = util.get_saml_configuration()
   mode = config.get(util.DASHBOARD_SECTION, util.MODE_KEY)
   if mode != 'management_station':
      return saml_settings
   rv = jobs.get_saml_custom_attributes(request.cookies)
   if not rv['ok']:
      return ({'error': 'custom attributes fetch failed'}, 500)
   saml_settings['saml_custom_attributes'] = rv['content']
   return saml_settings

@app.route('/api/v1/settings/auth/saml', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_settings_auth_saml_put():
   data = request.get_json()
   # admin auth saml
   admin_saml_provider = validators.get_enum('admin_saml_provider', required=False,
                                      values=['ADFS', 'Other', 'Siteminder',
                                              'ADFS_IWA', 'Azure', 'Otka']) or ''
   admin_saml_endpoint_url = validators.get_url('admin_saml_endpoint_url',
                                          required=False) or ''
   admin_saml_issuer_url = validators.get_string('admin_saml_issuer_url',
                                           required=False) or ''
   admin_saml_x509_certificate = validators.get_public_cert('admin_saml_x509_certificate',
                                                      required=False) or ''
   admin_saml_x509_certificate_additional = validators.get_public_cert('admin_saml_x509_certificate_additional',
                                                      required=False) or ''
   admin_saml_signed_requests = validators.get_bool('admin_saml_signed_requests',
                                                    required=False) or False
   admin_saml_static_entity_enabled = validators.get_bool('admin_saml_static_entity_enabled', required=False) or False

   if (not (admin_saml_endpoint_url and admin_saml_issuer_url and admin_saml_x509_certificate)
       and (admin_saml_endpoint_url or admin_saml_issuer_url or admin_saml_x509_certificate)):
      raise validators.ConflictingValuesError('Admin SAML settings')

   # service saml
   saml_provider = validators.get_enum('saml_provider', required=False,
                                      values=['ADFS', 'Other', 'Siteminder',
                                              'ADFS_IWA', 'Azure']) or ''
   saml_endpoint_url = validators.get_url('saml_endpoint_url',
                                          required=False) or ''
   saml_issuer_url = validators.get_string('saml_issuer_url',
                                           required=False) or ''
   saml_x509_certificate = validators.get_public_cert('saml_x509_certificate',
                                                      required=False) or ''
   saml_x509_certificate_additional = validators.get_public_cert('saml_x509_certificate_additional',
                                                      required=False) or ''
   saml_signed_requests = validators.get_bool('saml_signed_requests',
                                              required=False) or False
   saml_custom_attributes = data.get('saml_custom_attributes')
   mode = config.get(util.DASHBOARD_SECTION, util.MODE_KEY)
   is_policy_server_enabled = config.getboolean(util.NETWORKING_SECTION,
         util.POLICY_SERVER_ENABLED_KEY)
   # validate that all are set, or none (UI does not have a master enable
   # switch)
   if (not (saml_endpoint_url and saml_issuer_url and saml_x509_certificate)
       and (saml_endpoint_url or saml_issuer_url or saml_x509_certificate)):
      raise validators.ConflictingValuesError('User SAML settings')
   # Enable the group capability in the tenant settings
   # when ADFS SAML provider is selected.
   if (is_policy_server_enabled and
       mode in {'safeview', 'safeview_proxy'} and
       saml_provider in {'ADFS', 'ADFS_IWA', 'Siteminder', 'Azure'}):
      jobs.put_tenant_group_capabilities(request.cookies)
   # Saml custom attributes
   if mode == 'management_station' and isinstance(saml_custom_attributes, list):
      for attribute in data['saml_custom_attributes']:
         validators.get_string('display_name', data=attribute)
         validators.get_string('attribute_name', data=attribute)
         validators.get_optional_string('description', data=attribute, required=False)
         validators.get_bool('is_same', data=attribute)
      jobs.put_saml_custom_attributes(data['saml_custom_attributes'],
                                      request.cookies)
   util.set_saml_configuration(saml_provider,
                               saml_endpoint_url,
                               saml_issuer_url,
                               saml_x509_certificate,
                               saml_x509_certificate_additional,
                               saml_signed_requests,
                               admin_saml_provider,
                               admin_saml_endpoint_url,
                               admin_saml_issuer_url,
                               admin_saml_x509_certificate,
                               admin_saml_x509_certificate_additional,
                               admin_saml_signed_requests,
                               admin_saml_static_entity_enabled
                               )
   if mode in {'safeview', 'safeview_proxy'}:
      jobs.restart_auth_server.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/auth/users')
@render_as_json
@authenticate()
def api_v1_get_settings_auth_users():
   return util.get_user_auth()

@app.route('/api/v1/settings/auth/users', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_settings_auth_users_put():
   """ Set user authentication providers.
   Web users:
        allow_anonymous: False | True
   --------------------+-------+-----
   saml_enabled: False | LDAP  | Anon
                 True  | SAML  | N/A

   Email users:
   allow_anonymous_readonly:      False | True
   -----------------------------+-------+-----
   saml_enabled_readonly: False | LDAP  | Anon
                          True  | SAML  | N/A
   """
   kwargs = {
      'allow_anonymous': validators.get_bool('allow_anonymous'),
      'allow_anonymous_readonly': validators.get_bool(
         'allow_anonymous_readonly'),
      'id_user_from_header': validators.get_bool('id_user_from_header'),
      'saml_enabled': validators.get_bool('saml_enabled'),
      'saml_enabled_readonly': validators.get_bool('saml_enabled_readonly'),
      'email_reauth_enabled': validators.get_bool('email_reauth_enabled'),
      'web_reauth_enabled': validators.get_bool('web_reauth_enabled')
      }
   if kwargs['saml_enabled'] and kwargs['allow_anonymous']:
      raise validators.ConflictingValuesError('saml/anon')
   if kwargs['saml_enabled_readonly'] and kwargs['allow_anonymous_readonly']:
      raise validators.ConflictingValuesError('saml/anon readonly')
   # Currently the auth-server does not support anonymous web access and
   # authenticate email (read only) access.
   if kwargs['allow_anonymous'] and not kwargs['allow_anonymous_readonly']:
      validators.ConflictingValuesError(
         'Anonymous web with authenticated email')
   # until backend support for different auth method for email vs. web,
   # validate that saml and ldap are not both used
   # https://jira.surfcrew.com/browse/PNR-1719
   _ldap_enabled = not (kwargs['saml_enabled'] or kwargs['allow_anonymous'])
   _ldap_readonly = not (kwargs['saml_enabled_readonly'] or
                         kwargs['allow_anonymous_readonly'])
   if ((kwargs['saml_enabled'] and _ldap_readonly) or
       (kwargs['saml_enabled_readonly'] and _ldap_enabled)):
      raise validators.ConflictingValuesError(
         'SAML and LDAP used')
   if kwargs['saml_enabled']:
      kwargs['saml_user_auth_with_group'] = validators.get_bool(
         'saml_user_auth_with_group')
      if kwargs['saml_user_auth_with_group']:
         kwargs['saml_authorized_group'] = validators.get_string(
            'saml_authorized_group', required=False) or ''
      else:
         kwargs['saml_authorized_attribute_name'] = validators.get_string(
            'saml_authorized_attribute_name')
         kwargs['saml_authorized_attribute_value'] = validators.get_string(
            'saml_authorized_attribute_value')
   elif not kwargs['allow_anonymous']:
      # LDAP auth
      kwargs['ldap_authorized_group'] = validators.get_string(
         'ldap_authorized_group', required=False) or ''
      kwargs['ldap_recursive_group_check'] = validators.get_bool(
         'ldap_recursive_group_check', required=False) or False
   if kwargs['email_reauth_enabled']:
      kwargs['email_reauth_period'] = validators.get_int(
            'email_reauth_period', min=1, max=365)
   if kwargs['web_reauth_enabled']:
      kwargs['web_reauth_period'] = validators.get_int(
            'web_reauth_period', min=1, max=365)
   util.set_user_auth(**kwargs)
   util.reload_and_restart_monit('set-user-auth')
   jobs.restart_auth_server.delay()
   return { 'ok': True }

@app.route('/api/v1/settings/auth/admins')
@render_as_json
@authenticate()
def api_v1_get_settings_auth_admins():
   auth_settings = util.get_admin_auth()
   auth_settings.update(util.get_cookie_expiration())
   return auth_settings

@app.route('/api/v1/settings/auth/admins', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_settings_auth_admins_put():
   """ Set the authentication method for the dashboard admin """
   ldap_admin_group_name = ''
   saml_admin_group_name = ''
   saml_admin_attribute_name = ''
   saml_admin_attribute_value = ''
   saml_admin_auth_with_group = config.get(util.AUTHENTICATION_SECTION,
                                           util.SAML_ADMIN_AUTH_WITH_GROUP_KEY)
   use_ldap_for_admin = validators.get_bool('use_ldap_for_admin')
   cookie_expiration = max(validators.get_int('cookie_expiration', required=False), 0)
   if validators.get_bool('use_ldap_for_admin'):
      ldap_admin_group_name = validators.get_string('ldap_admin_group_name')
   use_saml_for_admin = validators.get_bool('use_saml_for_admin')
   if use_saml_for_admin:
      saml_admin_auth_with_group = validators.get_bool('saml_admin_auth_with_group')
      if saml_admin_auth_with_group:
         saml_admin_group_name = validators.get_string('saml_admin_group_name',
                                            required=False) or ''
      else:
         saml_admin_attribute_name = validators.get_string('saml_admin_attribute_name',
                                             required=False) or ''
         saml_admin_attribute_value = validators.get_string('saml_admin_attribute_value',
                                             required=False) or ''
   # only one of use_ldap_for_admin and use_saml_for_admin can be true
   if use_ldap_for_admin and use_saml_for_admin:
      raise validators.ConflictingValuesError(
         'use_ldap_for_admin/use_saml_for_admin')
   util.set_admin_auth(use_ldap_for_admin, ldap_admin_group_name,
                       use_saml_for_admin, saml_admin_group_name,
                       saml_admin_attribute_name, saml_admin_attribute_value,
                       saml_admin_auth_with_group)
   if cookie_expiration > 0:
      util.set_cookie_expiration(cookie_expiration)
   # No need to restart any service (changes only affect dashboard)
   return { 'ok': True }

@app.route('/api/v1/settings/ssl')
@render_as_json
@authenticate()
def api_v1_settings_ssl():
   return util.get_certificate_info()

# FIXME SECURITY: Add CSRF Token
@app.route('/api/v1/settings/ssl', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_post_settings_ssl():
   # Setting the "restarting" flag to have the web-app wait until the
   # certificate is loaded by director/haproxy and the dashboard service
   # restarts.
   filename = validators.get_certificate('certificate')
   validators.put_certificate(filename)
   return { 'ok': True }

@app.route('/api/v1/settings/ssl_bump')
@render_as_json
@authenticate()
def api_v1_settings_ssl_bump():
   return util.get_certificate_info(validators.PNR_SQUID_BUMP_KEY_CERT)

# FIXME SECURITY: Add CSRF Token
@app.route('/api/v1/settings/ssl_bump', methods=['POST'])
@render_as_json
@authenticate()
def api_v1_post_settings_ssl_bump():
   filename = validators.get_ca_certificate('certificate')
   validators.put_certificate(filename, validators.PNR_SQUID_BUMP_KEY_CERT,
                              'squid:squid')
   return { 'ok': True }

@app.route('/api/v1/settings/safemail', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_get_settings_safemail():
   return util.get_safemail_settings()

@app.route('/api/v1/settings/safemail', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_put_settings_safemail():
   direction = validators.get_enum('direction', values=['inbound', 'outbound'])
   hostname = validators.get_host('hostname')
   port = validators.get_port('port')
   safeview = validators.get_host('safeview')
   tls_enabled = validators.get_enum('tls_enabled',
                                     values=['always', 'never', 'opportunistic'])
   tls_validate = validators.get_bool('tls_validate')
   restrict_senders = validators.get_bool('restrict_senders')
   allowed_senders = validators.get_ip_ranges('allowed_senders',
                                              required=restrict_senders)
   # allowed_senders is optional. Make sure to normalize it to str
   if allowed_senders is None:
      allowed_senders = ''
   attachments_enabled = validators.get_bool('attachments_enabled')
   reversing_labs_enabled = validators.get_bool('reversing_labs_enabled')
   reversing_labs_url = ''
   if reversing_labs_enabled:
      reversing_labs_url = validators.get_url('reversing_labs_url')
   rv = util.set_safemail_settings(direction, hostname, port, safeview,
                                   tls_enabled, tls_validate,
                                   restrict_senders, allowed_senders,
                                   attachments_enabled, reversing_labs_enabled,
                                   reversing_labs_url)
   # restart safemail
   jobs.restart_safemail.delay()
   util.setup_iptables()
   return {'ok': rv}

@app.route('/api/v1/settings/attachments', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_get_settings_safemail_attachments():
   return util.get_safemail_attachment_settings()

@app.route('/api/v1/settings/attachments', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_put_settings_safemail_attachments():
   rl_enabled = validators.get_bool('reversing_labs_enabled')
   rl_host = validators.get_host_port('reversing_labs_host', required=rl_enabled) or ''
   rl_key = validators.get_optional_string('reversing_labs_key', required=False) or ''
   rv = util.set_safemail_attachment_settings(rl_enabled, rl_host, rl_key)
   # restart file-server, pnr-enforcement, etc.
   jobs.restart_safefile_processes.delay()
   return {'ok': rv}

@app.route('/api/v1/settings/all')
@render_as_json
@authenticate()
def api_v1_get_settings_all():
   d = {}
   settings.get_config_settings(d)
   return d

@app.route('/api/v1/settings/all', methods=['PUT'])
@render_as_json
@authenticate()
@restart_service
def api_v1_put_settings_all():
   new_settings = request.get_json()
   # validate template doesn't attempt to modify the system's mode
   if 'dashboard' in new_settings and 'mode' in new_settings['dashboard']:
      if new_settings['dashboard']['mode'] != config.get('dashboard', 'mode'):
         logging.warning('Wrong mode: %s (should be %s)',
                         new_settings['dashboard']['mode'],
                         config.get('dashboard', 'mode'))
         raise validators.InvalidValueError('dashboard.mode')
   jobs.apply_config_settings.delay(new_settings)
   return {'ok': True}

# FIXME SECURITY: For version 2.0 we should reimplement this without flask and
# use Tornado's CSRF check
@app.route('/api/v1/settings_export', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_post_settings_export():
   data = request.get_json()
   password = data.get("password")
   job = jobs.export_settings.delay(password=password, type="settings", filename="menlo-settings")
   return { 'id': job.id }

# FIXME SECURITY: Need to add a CSRF token to this post method
@app.route('/api/v1/settings_import', methods=["POST"])
@render_as_json
@authenticate()
@restart_service
def api_v1_post_settings_import():
   try:
      filename = validators.get_imported_settings("settings")
      # Apply imported settings file here in order to be able to propagate
      # issues to the UI. Currently this is a quick process
      # Need the cookies to communicate with pnr-policy
      settings.import_settings_from_json_file(filename, request.cookies)
      # Trigger same restart as template apply
      jobs.reload_and_restart_all.delay()
      return { 'ok': True }
   except Exception as err:
      err_str = str(err)
      if err_str == 'No JSON object could be decoded':
         err_str = 'Invalid file or wrong password'
      raise validators.ValidationError('settings', 'import-error', err_str)

# FIXME SECURITY: Add CSRF Token
@app.route('/api/v1/support')
@authenticate(html=True)
def api_v1_support():
   enable = request.args.get('enable')
   if not enable:
      return 'Enable/disable support mode with ?enable=true|false'
   if enable in [ '1', 'true', 'on' ]:
      util.enable_support(True)
      return 'Support mode enabled'
   util.enable_support(False)
   return 'Support mode disabled'

@app.route('/api/v1/restrict_ssl_tls')
@authenticate(html=True)
def api_v1_restrict_ssl_tls():
   inbound = request.args.get('inbound')
   outbound = request.args.get('outbound')
   return_msg = []

   if not (inbound or outbound):
      return 'Enable/disable restrictive SSL-TLS support with ?inbound=true|false&outbound=true|false'

   if inbound:
      restrict_inbound = inbound.lower() in ['1', 'true', 'on']
      util.enable_restrictive_ssl_tls(util.TLS_MIN_VERSION_INBOUND_KEY, restrict_inbound)
      return_msg.append('Inbound restrictive SSL-TLS support %s.' % ('enabled' if restrict_inbound
                                                                     else 'disabled'))

   if outbound:
      restrict_outbound = outbound.lower() in ['1', 'true', 'on']
      util.enable_restrictive_ssl_tls(util.TLS_MIN_VERSION_OUTBOUND_KEY, restrict_outbound)
      return_msg.append('Outbound restrictive SSL-TLS support %s.' % ('enabled' if restrict_outbound
                                                                      else 'disabled'))

   jobs.upstart_restart_all.delay()
   return ' '.join(return_msg)

@app.route('/api/v1/plugins')
@authenticate(html=True)
def api_v1_plugins():
   try:
      enable = validators.get_enum('enable',
                                   values={'false', 'flash', 'java', 'all'},
                                   data=request.args)
      global_flash_on = validators.get_enum('global_flash_on',
                                            values={'true', 'false'},
                                            data=request.args)
   except (validators.RequiredValueError, validators.InvalidValueError):
      return 'Enable/disable plugin support with ?enable=false|flash|java|all&global_flash_on=true|false'
   rv = util.set_plugin_support(enable) and util.set_global_flash_on(global_flash_on)
   if rv:
      jobs.restart_surrogates.delay()
   return 'Plugins enable = %s, global_flash_on = %s' % (enable, global_flash_on)

@app.route('/api/v1/compat')
@authenticate(html=True)
def api_v1_compat():
   enable = request.args.get('enable')
   if not enable:
      return 'Set browser compatibility with ?enable=default|IE:9|[custom]'
   if enable == 'default':
      enable = 'Android:4.1,Firefox:8,Mobile Safari:5,Chrome Mobile:18,Chrome Mobile iOS:18,Safari:5,Chrome:12,Chromium:12,IE:9'
   rv = util.set_compat_list(enable)
   if rv:
      jobs.restart_surrogates.delay()
   return 'Browser compatibility = %s' % enable

@app.route('/api/v1/reset', methods=["PUT"])
@render_as_json
@authenticate()
@restart_service
def api_v1_reset():
   app.reporter.report('admin_reset_service', {})
   jobs.upstart_restart_all.delay()
   return { 'ok': True }

@app.route('/api/v1/reboot', methods=["PUT"])
@render_as_json
@authenticate()
@restart_service
def api_v1_reboot():
   app.reporter.report('admin_reboot_machine', {})
   app.reporter.report('safeview_down', {
      'status_desc': 'Admin initiated a reboot'
   })
   jobs.reboot_system()
   return { 'ok': True }

@app.route('/api/v1/shutdown', methods=["PUT"])
@render_as_json
@authenticate()
@restart_service
def api_v1_shutdown():
   app.reporter.report('admin_shutdown_machine', {})
   app.reporter.report('safeview_down', {
      'status_desc': 'Admin initiated a shutdown'
   })
   jobs.shutdown_system()
   return { 'ok': True }

@app.route('/api/v1/maintenance_mode', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_maintenance_mode_get():
   return {'ok': True,
           'enabled': util.get_maintenance_mode()}

@app.route('/api/v1/maintenance_mode', methods=['PUT'])
@appliance_only_route
@render_as_json
@authenticate()
def api_v1_maintenance_mode_put():
   enable = validators.get_bool('enable')
   action_str = 'enable' if enable else 'disable'
   util.set_maintenance_mode(enable)
   jobs.apply_maintenance_mode.delay()
   app.reporter.report('admin_%s_maintenance_mode' % action_str,
                       {'user_id': _get_user_name()})
   return {'ok': True}

@app.route('/api/v1/password', methods=["POST"])
@render_as_json
@authenticate(chpass=False)
def api_v1_password():
   data = request.get_json()
   old_passwd = data.get("old_password")
   new_passwd = data.get("new_password")

   if util.is_first_login():
      old_passwd = DEFAULT_PASSWORD
   if not old_passwd or not new_passwd or new_passwd == DEFAULT_PASSWORD:
      return (Error.INVALID_PASSWORD, 400)
   if not pam.authenticate(ADMIN_USER, old_passwd):
      return (Error.INVALID_PASSWORD, 400)
   if not util.change_password(ADMIN_USER, old_passwd, new_passwd):
      return (Error.INVALID_PASSWORD, 400)
   if util.is_first_login():
      util.unset_first_login()

   app.reporter.report('admin_change_password', {})

   # Regnerate the token timestamp which will invalidate all tokens that are
   # currently in use
   _globals.api_token_valid_after = make_timestamp()
   db.set_token_valid_after(_globals.api_token_valid_after)
   token = Token.make(app, ADMIN_USER)
   session['token'] = token
   return { 'ok': True, 'token': token }

@app.route('/api/v1/exports')
@render_as_json
@authenticate()
def api_v1_get_exports():
   logs_q = rq.Queue(jobs.DASHBOARD_LOGS_Q)

   queued_jobs = [ jobs.get_job_status(job) for job in logs_q.jobs ]
   active_jobs = [ jobs.get_job_status(w.get_current_job())
      for w in rq.Worker.all() ]
   return {
      'jobs': queued_jobs + [x for x in active_jobs if x]
   }

# FIXME SECURITY: Need to add a CSRF token to this post method
@app.route('/api/v1/logs', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_post_logs():
   app.reporter.report('admin_export_logs', {})
   job = jobs.export_logs.delay(type="logs", filename="logs.tar.bz2")
   return { 'id': job.id }

# FIXME SECURITY: Need to add a CSRF token to this post method
@app.route('/api/v1/diagnostics', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_post_diagnostics():
   job = jobs.export_diagnostics.delay(type="diagnostics",
         filename="diagnostics.tar.bz2")
   return { 'id': job.id }

@app.route('/api/v1/exports/<job_id>/status')
@render_as_json
@authenticate()
def api_v1_exports_status(job_id):
   job = rq.Queue(jobs.DASHBOARD_LOGS_Q).fetch_job(job_id)
   if job:
      return jobs.get_job_status(job)
   return (Error.JOB_NOT_FOUND, 404)

# {kowsik} Verified that the send_file (when passed in a file object)
# closes the file descriptor after the transfer. In addition we also unlink
# the file so there's no left overs
@app.route('/api/v1/exports/<job_id>')
@authenticate(html=True)
def api_v1_exports(job_id):
   job = rq.Queue(jobs.DASHBOARD_LOGS_Q).fetch_job(job_id)
   if job and job.is_finished:
      tmpfname = job.result
      tmpf = open(tmpfname, "r")
      os.unlink(tmpfname)
      kwargs = {
         'mimetype': 'application/octet-stream',
         'as_attachment': True,
         'attachment_filename': job.kwargs['filename'],
         'add_etags': False
      }
      return send_file(tmpf, **kwargs)
   return ''


@app.route('/api/v1/logs/upload', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_upload_diagnostics():
   data = request.get_json()
   job = jobs.upload_logs.delay(data.get('ticket'))
   return { 'id': job.id }

@app.route('/api/v1/update')
@render_as_json
@authenticate()
def api_v1_get_update():
   return jobs.get_update_info()

# FIXME SECURITY: Need to add a CSRF token to this post method
@app.route('/api/v1/update', methods=["POST"])
@render_as_json
@authenticate()
def api_v1_post_update():
   validators.check_if_update_allowed()
   val = validators.get_software_update('update')
   return val

@app.route('/api/v1/update/check', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_put_update_check():
   job = jobs.check_for_update.delay(any_ver=request.args.get('any'))
   return { 'id': job.id }

@app.route('/api/v1/update/check/<job_id>')
@render_as_json
@authenticate()
def api_v1_get_update_check_id(job_id):
   return _get_job_status_result(job_id, jobs.DASHBOARD_UPDATE_Q)

# If an online update is available and if it hasn't been downloaded already,
# then schedule a background job to download this.
@app.route('/api/v1/update/online', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_put_update_online():
   validators.check_if_update_allowed()
   job = jobs.download_update.delay(any_ver=request.args.get('any'))
   update_id = 'online:' + job.id
   return { 'ok': True, 'id': update_id, 'status': 'queued',
      'source': 'online' }

@app.route('/api/v1/update/<_id>', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_put_update(_id):
   _globals.upgrade_token = upgrade.prepare_upgrade()
   source, dummy_basename = validators.decode_update_id(_id)
   if request.data and validators.get_bool('enter_maintenance_mode',
                                           required=False):
      util.set_maintenance_mode(True)
      jobs.apply_maintenance_mode.delay()
      app.reporter.report('admin_enable_maintenance_mode',
                          {'user_id': _get_user_name()})
   if source == 'online':
      _globals.upgrading = True
   job = jobs.install_update.delay(_id, _globals.upgrade_token)
   return {'id': job.id, 'source': source}

@app.route('/api/v1/update/<_id>', methods=["DELETE"])
@render_as_json
@authenticate()
def api_v1_delete_update(_id):
   validators.del_software_update(_id)
   return { 'ok': True }

@app.route('/api/v1/update/auto_check', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_update_auto_check_get():
   return util.get_auto_update_check_configuration()

@app.route('/api/v1/update/auto_check', methods=['PUT'])
@render_as_json
@authenticate()
def api_v1_update_auto_check_put():
   enabled = validators.get_bool(util.AUTO_UPDATE_CHECK)
   util.set_auto_update_check_configuration(enabled)
   return { 'ok' : True, util.AUTO_UPDATE_CHECK : enabled }

@app.route('/api/v1/update/rollback', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_update_rollback_get():
   return {'ok': True, 'rollbacks': upgrade.get_rollback_images()}

@app.route('/api/v1/update/rollback/<slot_id>', methods=['PUT'])
@render_as_json
@authenticate()
@restart_service
def api_v1_update_rollback_put(slot_id):
   # TODO: validate the slot ID (what is the criteria?)
   version = None
   for i in upgrade.get_system_images():
      if i['slot_id'] == slot_id:
         version = i['version']
         break
   if version is None:
      logging.error('Could not find slot %s', slot_id)
      raise validators.InvalidRollbackError
   rv = upgrade.perform_rollback(slot_id)
   if rv:
      app.reporter.report('admin_software_rollback', {'version': version})
      jobs.reboot_system()
   return {'ok': rv}

@app.route('/api/v1/update/reset', methods=['PUT'])
@render_as_json
@authenticate()
@restart_service
def api_v1_update_reset_put():
   rv = upgrade.perform_factory_reset()
   if rv:
      app.reporter.report('admin_software_reset', {})
      jobs.reboot_system()
   return {'ok': rv}

@app.route('/api/v1/update/min_version', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_get_min_update_version():
   return {
      'central_upgradeable': config.get('upgrades',
                                        'min_central_upgradeable_version'),
      'peer_download': config.get('upgrades',
                                  'min_peer_download_version')
   }

@app.route('/api/v1/admin/disable')
@authenticate(html=True)
def api_v1_get_admin_disable():
   success = util.change_admin_lock(False)

   if not success:
      return ('Unable to disable admin account', 400)

   app.reporter.report('admin_account_disabled', {})
   return 'Admin account successfully disabled'

@app.route('/api/v1/get_client_info')
@appliance_only_route
@render_as_json
@authenticate()
def api_v1_get_client_info():
   info = {db.UUID: trust.uuid(),
           db.HOSTNAME: net_utils.get_mgmt_hostname(),
           db.CLIENT_CERT: open(trust.MY_CLIENT_CERT).read(),
           db.SERVICE_DOMAIN: config.get('service', 'service_domain'),
           db.MODE: config.get('dashboard', 'mode'),
           db.BUILD_ID: get_build_id(),
           # It is currently assumed that these will not change for the life
           # of a system.
           db.ROLES: config.get('instance-userdata', 'roles'),
           db.NR_CPUS:  psutil.cpu_count()}
   return info

_status_timeout = None

# Undocumented APIs to get/set the current status of DB
@app.route('/api/v1/db_status', methods=['GET'])
@render_as_json
def api_v1_db_status_get():
   return db.get_db_status(all_fields=True)

@app.route('/api/v1/db_status', methods=['PUT'])
@render_as_json
def api_v1_db_status_clear():
   info = {'status': 'ok', 'details': '', 'message': ''}
   try:
      db.set_db_status(**info)
      return {'ok': True}
   except Exception:
      return {'ok': False, 'reason': 'backend error'}

_license_check_timeout = None

def _license_check():
   """Periodically check the supplied license string."""
   global _license_check_timeout

   try:
      jobs.check_license_features.delay()
   except Exception as e:
      # Catch all exceptions
      logging.exception('event="periodic-license-check-failed" error="%s" '
                        'error_type="%s"', e, type(e).__name__)
   finally:
      _license_check_timeout = IOLoop.current().call_later(
         config.getint('dashboard', 'license_check_period') +
         random.randint(
            0, config.getint(
               'dashboard', 'license_check_period_random_window')),
         _license_check)

def _start_license_check():
   active_licensed_features = [
      feature for feature, enabled
      in util.licensed_features_current_setting().iteritems() if enabled]
   util.store_active_licensed_features(active_licensed_features)
   util.store_expected_licensed_features(active_licensed_features)
   logging.info('event="startup-licensed-features" features="%s"',
                active_licensed_features)
   logging.info('Scheduling license checks')
   _license_check()

@app.route('/api/v1/app_config.js')
def api_v1_app_config_js():
   """ return javascript file containing web app's configuration """
   app_conf = {
         'mode': config.get('dashboard', 'mode'),
         'stats_period': config.getint('dashboard', 'stats_period'),
         }
   return 'var APP_CONFIG = %s;' % json.dumps(app_conf)

@app.route('/api/v1/learning')
@authenticate(html=True)
def api_v1_learning():
   """ DEPRECATED """
   enable = request.args.get('enable')
   if not enable:
      return 'Enable/disable readonly learning mode with ?enable=true|false'
   if enable in [ '1', 'true', 'on' ]:
      util.enable_readonly_learning(True)
      return 'Readonly learning mode enabled'
   util.enable_readonly_learning(False)
   return 'Readonly learning mode disabled'

@app.route('/api/v1/anonymous_readonly')
@authenticate(html=True)
def api_v1_anonymous_readonly():
   """ DEPRECATED Enable or disable anonymous readonly mode """
   enable = request.args.get('enable')
   if not enable:
      return 'Enable/disable anonymous access to readonly mode with ?enable=true|false'
   if enable in [ '1', 'true', 'on' ]:
      util.enable_anonymous_readonly(True)
      return 'Anonymous readonly mode enabled'
   util.enable_anonymous_readonly(False)
   return 'Anonymous readonly mode disabled'

@app.route('/api/v1/document_isolation')
@authenticate(html=True)
def api_v1_document_isolation():
   """ Enable or disable document isolation mode """
   enable = request.args.get('enable')
   if not enable:
      return 'Enable/disable document isolation with ?enable=true|false'
   if enable in [ '1', 'true', 'on' ]:
      util.enable_safedocs(True)
      return 'Document Isolation enabled'
   util.enable_safedocs(False)
   return 'Document Isolation disabled'

@app.route('/api/v1/mx_rr')
@authenticate(html=True)
def api_v1_mx_rr():
   enable = request.args.get('enable')
   if not enable:
      return 'Enable/disable DNS MX Round-robin with ?enable=true|false'
   if enable in [ '1', 'true', 'on' ]:
      util.enable_mx_rr(True)
      return 'DNS MX RR enabled'
   util.enable_mx_rr(False)
   return 'DNS MX RR disabled'

@app.route('/api/v1/risk_score')
@authenticate(html=True)
def api_v1_risk_score():
   _range = request.args.get('range')
   r = (_range or ',').split(',',1)
   if not(len(r) == 2 and r[0].isdigit() and r[1].isdigit()):
      return ('Specify integer medium and high Risk Score to set thresholds in '
              '?range=medium,high <br>The current thresholds are: ' +
              util.get_risk_score())
   med = int(r[0])
   high = int(r[1])
   return util.set_risk_score(med, high)

@app.route('/api/mgr/v1/get_sv_cert/<hostname>/<int:port>')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_get_sv_cert(hostname, port):
   job = jobs.get_server_cert_details.delay(hostname, port)
   return { 'id': job.id }

@app.route('/api/mgr/v1/groups')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_get_all():
   """ called by the web app to get group management info
       returns all the groups and associated metadata """
   return make_response(json.dumps(db.get_group_info()), 200,
         {'Content-Type': 'application/json'})

@app.route('/api/mgr/v1/groups/<int:group_id>')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_get_specific(group_id):
   """ called by the web app to get group management info
       returns the metadata for a specific group requested """
   return make_response(json.dumps(db.get_group_info(group_id)), 200,
         {'Content-Type': 'application/json'})

def _validate_group_info(group_info, op_create=False):
   try:
      validators.get_int(db.GROUP_ID, data=group_info, required=False)
      if op_create:
         validators.get_string(db.GROUP_NAME, data=group_info)
         validators.get_enum(db.GROUP_TYPE, data=group_info, values=[
               db.GROUP_TYPE_CLUSTER, db.GROUP_TYPE_STANDALONE
            ])
      validators.get_string(db.GROUP_DESC, data=group_info, required=False)
      validators.get_ip(db.GROUP_STORAGE_IP, data=group_info, required=False)
      validators.get_string(db.GROUP_STORAGE_LOCATION, data=group_info,
                            required=False)
      # TODO: check if group type is valid.
      return {'ok': True}
   except validators.InvalidValueError as ex:
      return {'ok': False, 'reason': ex.message}
   except validators.RequiredValueError as ex:
      return {'ok': False, 'reason': ex.message}

@app.route('/api/mgr/v1/groups', methods=['PUT'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_create():
   """ called by the web app to create a new group """
   error = {'reason': ''}
   group_info = request.get_json()
   rv = _validate_group_info(group_info, True)
   if not rv['ok']:
      error['reason'] = rv['reason']
      return (error, 400)
   return db.group_create(group_info)

@app.route('/api/mgr/v1/groups/<int:group_id>', methods=['POST'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_update(group_id):
   """ called by the web app to update an existing group """
   error = {'reason': ''}
   group_info = request.get_json()
   rv = _validate_group_info(group_info, False)
   if not rv['ok']:
      error['reason'] = rv['reason']
      return (error, 400)
   return db.group_update(group_id, group_info)

@app.route('/api/mgr/v1/groups/<int:group_id>', methods=['DELETE'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_delete(group_id):
   """ called by the web app to delete an existing group """
   # Will not be deleted if group is non-empty.
   return db.group_delete(group_id)

@app.route('/api/mgr/v1/groups/move/<int:new_group_id>/<int:old_group_id>/',
           methods=['POST'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_groups_move_node(new_group_id, old_group_id):
   """ called by the web app to add node to an existing group """
   error = {'reason': ''}
   data = request.get_json()
   nodes = data.get('nodes')
   if not isinstance(nodes, 'list'):
      error['reason'] = 'Invalid format for nodes'
      return (error, 400)
   return db.do_node_operation(db.NODE_MOVE, new_group_id, nodes,
                               old_group_id)

# DEPRECATED in 2.57.1
@app.route('/api/mgr/v1/approve_sv/<hostname>/<int:port>', methods=['POST'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_approve_sv(hostname, port):
   data = request.get_json()
   username = data.get('username', None)
   password = data.get('password', None)
   cert = data.get('cert')
   name = data.get('name')
   # Used for two step finalization of trust and adding appliance.
   # Step-1: Get basic appliance info, return the group information
   # Step-2: Get group information, create group if required, and add
   #         the appliance to the group and finalize trust.
   finalize_trust = data.get('finalize_trust', True)
   # If a group id is not sent, fallback to 'unassigned'
   group_id = data.get('group_id', db.UNASSIGNED_GROUP_ID)
   job = jobs.build_trust.delay(hostname, port, cert, username, password,
                                name, finalize_trust, group_id)
   return {'id': job.id}

@app.route('/api/mgr/v1/jobs/<job_id>/status')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_get_job_status(job_id):
   # TODO: maybe use a different queue for manager jobs in the future
   return _get_job_status_result(job_id)

@app.route('/api/mgr/v1/cluster_info', methods=['GET'])
@manager_route
@render_as_json
@authenticate(cert_required=True)
def api_mgr_v1_cluster_info():
   """ called by a managed safeview-proxy node to discover cluster members """
   # FIXME: the render_as_json decorator does special handling of lists and
   # tuples
   return make_response(
      json.dumps(db.get_sv_cluster_info(request.headers.get('X-SSL-Client-CN'))),
      200,
      {'Content-Type': 'application/json'})

# TODO: Deprecated since 2.53. New functionality in the v2 route in
# dashboard-ng. Kept for backwards compatibility.
@app.route('/api/mgr/v1/status', methods=['POST'])
@manager_route
@render_as_json
@authenticate(cert_required=True)
def api_mgr_v1_status_post():
   """ called by a managed safeview system to post its status """
   # FIXME: pylint error: https://jira.surfcrew.com/browse/PNR-2802
   # pylint: disable=too-many-branches
   result = {}
   uuid = request.headers.get('X-SSL-Client-CN')
   if db.get_safeview_trust_state(uuid) != db.TRUSTED:
      logging.warning('dropping status from untrusted system %s', uuid)
      return {'ok': False}
   data = request.get_json()
   # Check to see if the service domain changed.
   saved_sd = db.get_safeview_service_domain(uuid)
   live_sd = data.get('settings', {}).get('service', {}).get('service_domain')
   if live_sd and live_sd != saved_sd:
      logging.info('service domain changed for %s from %s to %s',
                   uuid, saved_sd, live_sd)
      db.set_safeview_service_domain(uuid, live_sd)
   # Update the internal IP (used to communicate between master/worker in
   # a cluster configuration) and the appliance version.
   # FIXME: Make sure that the connection version is same before getting
   # the internal ip address.
   if 'internal_ip' in data:
      info = {'uuid': uuid, 'internal_ip': data['internal_ip']}
      if data.get('has_build_id_support'):
         info['build_id'] = data['build']
      db.upsert_safeview(info)
   else:
      logging.warning('%s did not send internal IP address', uuid)
   # Verify settings are up to date with configuration template (if applied)
   template = db.get_sv_config_template(uuid)
   if template:
      diffs = settings.check_managed_settings(data.get('settings', {}),
                                              template[db.TEMPLATE])
      if diffs:
         # catch the corner case where a new status message was received after
         # the template apply and before the system restart.
         try:
            last_apply = float(template[db.TEMPLATE_APPLY_TS])
         except (ValueError, TypeError):
            last_apply = 0
         # Are changes still being propagated to the instance?
         diffs = 'pending' if (
            time.time() - last_apply < db.APPLY_PENDING_TIMEOUT) else diffs
      data['settings_state'] = diffs
   # Check if the connection is up to date
   if (not data.get(db.CONNECTION_VERSION) or
       int(data.get(db.CONNECTION_VERSION)) != trust.CONNECTION_VERSION):
      # connection is out of date
      if (not data.get('trust_version') or
          int(data.get('trust_version')) != trust.CONNECTION_VERSION):
         # Managed system should be updated
         # TODO: Need to surface this to the admin
         logging.warning('system %s should be upgraded', uuid)
      else:
         # TODO: update the connection
         pass
   else:
      # Connection is up to date, can proceed with other checks
      # Check logging state
      if data[util.FLUENTD_ENABLED] != config.getboolean(
            util.PNR_DRUID_SECTION, util.DRUID_LOGGING_ENABLE_KEY):
         logging.info('need to update managed system logging')
         trust.add_logging_config(result)

      # Get group id and pass the cookie secret
      # FIXME: only for cluster nodes as of now
      # standalone: (JIRA: PNR-2593)
      group_id = db.get_safeview_node_group_id(uuid)
      if db.get_group_type(group_id) == db.GROUP_TYPE_CLUSTER:
         cookie_secret = db.get_group_cookie_secret(group_id)
         if cookie_secret:
            result['cookie_secret'] = cookie_secret

      # If the safefile hash has changed, send the new config
      if (data.get('safefile.scp_hash') and
          data['safefile.scp_hash'] != config.get(
            util.SAFEFILE_SECTION, util.SAFEFILE_SCP_HASH)):
         logging.info('need to update av file scan config')
         util.get_safefile_credentials(result)
   # Store state in DB
   result['ok'] = db.update_sv_status(uuid, data)
   return result
   # pylint: enable=too-many-branches

def _get_cmr_status(is_secondary_cmr=False):
   status = util.get_status()
   status.update(util.get_product_info())
   return {
      'name': db.SEC_CMR_HOSTNAME if is_secondary_cmr else 'Central Manager',
      'service_domain': 'managers',
      'data': [status],
      'uuid': trust.uuid(),
      'addr': get_bind_address('management'),
      'port': config.getint('dashboard', 'cam_port'),
      'trust_state': 'trusted',
      'is_manager': True,
      'group': 'managers',
      'build_id': get_build_id(),
       # TODO: UI has to decode this accordingly and display a 'red' status if
       # something is wrong.
      'db_status': db.get_db_status()
   }

@app.route('/api/mgr/v1/status')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_status_get():
   """ called by the web app to get centralized status """
   try:
      limit = max(0, int(request.args.get('limit', 0)))
   except ValueError:
      limit = 0
   is_secondary_cmr = util.is_secondary_cmr()
   my_uuid = trust.uuid() if is_secondary_cmr else None
   statuses = db.get_sv_statuses(limit, my_uuid=my_uuid)
   if util.is_cmr():
      statuses.append(_get_cmr_status(is_secondary_cmr))
   return {'statuses': statuses, 'last_seen': db.get_sv_last_update_times()}

@app.route('/api/mgr/v1/managed_systems')
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_managed_systems_get():
   """ called by the web app to get details of managed systems """
   my_uuid = trust.uuid() if util.is_secondary_cmr() else None
   return {'managed_systems': db.get_all_safeview_details(
      util.local_admin_disabled(),
      my_uuid=my_uuid
   )}

@app.route('/api/mgr/v1/managed_systems/<uuid>', methods=['DELETE'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_managed_systems_delete(uuid):
   """ called by the web app to stop managing a system """
   job = jobs.trust_delete_safeview.delay(uuid)
   return { 'id': job.id }

@app.route('/api/mgr/v1/current_settings/<sv_uuid>', methods=['GET'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_current_settings_get(sv_uuid):
   """ Get the current settings from the running instance """
   sv_addr, sv_port = db.get_safeview_addr(sv_uuid)
   job = jobs.get_settings.delay(sv_addr, sv_port)
   return { 'id': job.id }

@app.route('/api/mgr/v1/templates', methods=['GET'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_templates_get():
   templates = db.get_all_sv_config_templates()
   return {'ok': True, 'templates': templates}

@app.route('/api/mgr/v1/template/<int:template_id>', methods=['GET'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_template_get(template_id):
   return db.get_config_template(template_id)

@app.route('/api/mgr/v1/template', methods=['POST'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_template_post_new():
   data = request.get_json()
   clean, invalid = settings.sanitize_template(data[db.TEMPLATE])
   if invalid:
      logging.warning('Found invalid template properties: %s', json.dumps(invalid))
      return (invalid, httplib.BAD_REQUEST)
   tid = db.save_config_template(clean, data[db.NAME])
   app.reporter.report('cam_create_template', {'user_id': _get_user_name(),
                                               'template_id': tid})
   return {
         'ok': True,
         'template_id': tid
         }

@app.route('/api/mgr/v1/template/<int:template_id>', methods=['POST'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_template_post(template_id):
   data = request.get_json()
   clean, invalid = settings.sanitize_template(data[db.TEMPLATE])
   if invalid:
      logging.warning('Found invalid template properties: %s', json.dumps(invalid))
      return (invalid, httplib.BAD_REQUEST)
   tid = db.save_config_template(clean, data[db.NAME], template_id)
   app.reporter.report('cam_modify_template', {'user_id': _get_user_name(),
                                               'template_id': tid})
   return {
         'ok': True,
         'template_id': template_id
         }

@app.route('/api/mgr/v1/template/<int:template_id>', methods=['DELETE'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_template_delete(template_id):
   rv = db.delete_config_template(template_id)
   if rv['ok']:
      app.reporter.report('cam_delete_template', {'user_id': _get_user_name(),
                                                  'template_id': template_id})
   return rv

@app.route('/api/mgr/v1/assign_templates', methods=['PUT'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_assign_templates_put():
   template_map = request.get_json()
   if not template_map:
      return (Error.MISSING_DATA, 400)
   result_map = {}
   report_map = {}
   ok = True
   for sv_uuid, template_id in template_map.iteritems():
      rv = db.assign_template(sv_uuid, template_id)
      if not rv['ok']:
         ok = False
      result_map[sv_uuid] = rv
      report_map[sv_uuid] = {'template_id': template_id,
                             'ok': rv['ok']}
   app.reporter.report('cam_assign_templates', {'user_id': _get_user_name(),
                                                'details': report_map})
   return {'ok': ok, 'details': result_map}

@app.route('/api/mgr/v1/apply_templates', methods=['PUT'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_apply_templates_put():
   template_map = request.get_json()
   jobs_map = {}
   for sv_uuid, template_id in template_map.iteritems():
      job = jobs.push_settings.delay(sv_uuid, template_id)
      jobs_map[sv_uuid] = job.id
   app.reporter.report('cam_publish_templates', {'user_id': _get_user_name(),
                                                 'details': template_map})
   return {'ok': True, 'jobs': jobs_map}

@app.route('/api/mgr/v1/disable_local_admin', methods=['PUT'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_disable_local_admin_put():
   targets = request.get_json()
   job = jobs.push_disable_local_admin.delay(targets)
   return {'id': job.id}

@app.route('/api/mgr/v1/reg', methods=['GET'])
@manager_route
@render_as_json
@authenticate()
def api_mgr_v1_reg_get():
   return {'ok': True,
           'secret_set': secret_util.is_secret_present('dashboard',
                                                       'trust_secret'),
           'auto_group_enabled': util.get_auto_group_setting()}

@app.route('/api/mgr/v1/reg', methods=['PUT'])
@manager_route
@render_as_json
@authenticate()
@restart_service
def api_mgr_v1_reg_put():
   data = request.get_json()
   secret = data.get('secret', '')
   if not secret:
      secret_util.set_secret('dashboard', 'trust_secret', '')
   elif secret != 'secret':
      secret_util.set_secret('dashboard', 'trust_secret',
                             validators.get_string('secret'))
   util.set_auto_group_setting(validators.get_bool('auto_group_enabled'))
   jobs.restart_dashboard_all()
   return {'ok': True}

@app.route('/api/mgr/v1/trust_step1', methods=['PUT'])
@manager_route
@render_as_json
def api_mgr_v1_trust_step1():
   uuid = validators.get_string('uuid')
   nonce = trust.step1(uuid)
   if nonce is None:
      return ({'error': 'invalid_request', 'reason': 'Invalid request'}, 401)
   else:
      logging.info('Step1: issuing challenge to %s', uuid)
      return {'ok': True, 'challenge': nonce}

@app.route('/api/mgr/v1/trust_step2', methods=['PUT'])
@manager_route
@render_as_json
def api_mgr_v1_trust_step2():
   uuid = validators.get_string('uuid')
   response = validators.get_string('response')
   mode = validators.get_enum('mode', values=[
      util.SAFEMAIL_MODE,
      util.SAFEVIEW_MODE,
      util.SAFEVIEW_PROXY_MODE,
      util.SAFEVIEW_ISOLATION_MODE,
      util.SAFEMAIL_WORKER_MODE])
   hostname = validators.get_host('hostname')
   port = validators.get_int('port')
   service_domain = validators.get_host('service_domain')
   appl_client_cert = validators.get_string('client_cert')
   appl_server_cert = validators.get_string('server_cert')
   challenge = validators.get_string('challenge')
   is_cluster_node = mode in {util.SAFEVIEW_PROXY_MODE,
                              util.SAFEVIEW_ISOLATION_MODE}
   group_name = validators.get_string('group_name', required=is_cluster_node)
   group_type = db.GROUP_TYPE_CLUSTER if is_cluster_node else db.GROUP_TYPE_ANY
   group_id = None
   if group_name:
      group_id = db.get_group_id_by_name(group_name, group_type)
      if group_id is None:
         raise validators.InvalidValueError(
            'group_name: %s of type %s not found' % (group_name, group_type))
   logging.info('processing step2 for %s', uuid)
   rv = trust.step2(uuid, response, mode, hostname, port, service_domain,
                    appl_client_cert, appl_server_cert, challenge, group_id)
   if rv.get('ok'):
      return rv
   return (rv, 401)

@app.route('/api/v1/settings/access', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_get_settings_acl():
   return util.get_acl_settings()

@app.route('/api/v1/settings/access', methods=['PUT'])
@render_as_json
@authenticate()
@restart_service
def api_v1_put_settings_acl():
   data = request.get_json()
   acl_settings = {}
   for service in data.keys():
      if service == 'snmpd':
         continue
      service_conf = {}
      service_conf['access']  = validators.get_bool('allow_all',
                                                    data=data[service],
                                                    required=True)
      if not service_conf['access']:
         service_conf['ip_range'] = ','.join([validators.get_acl_ranges('text',
                                              data=value) for value in
                                              data[service]['ip_range']])
      acl_settings[service] = service_conf
   # SSH access ("remote CLI")
   allow_ssh = validators.get_bool('allow_ssh',
                                   data=data.get('management', {}),
                                   required=False)
   # Ping (ICMP) access
   allow_icmp = validators.get_bool('allow_icmp',
                                    data=data.get('management', {}),
                                    required=False)

   # Internal network access via MSIP
   allow_private_nets = validators.get_bool(util.ALLOW_PRIVATE_NETS_KEY,
                                        data=data.get('management', {}),
                                        required=False)
   snmpd = data.get('snmpd')
   if snmpd:
      acl_settings['snmpd'] = {}
      acl_settings['snmpd']['enabled']= validators.get_bool('enabled', data=data['snmpd'], required=False)
      acl_settings['snmpd']['ro_community']= validators.get_string('ro_community', data=data['snmpd'], required=False)
   rv = util.set_acl_settings(acl_settings, allow_ssh, allow_icmp, allow_private_nets)
   # reconfigure iptables
   util.setup_iptables()
   jobs.reload_and_restart_all.delay()
   return {'ok': rv}

@app.route('/api/v1/settings/scp_public_key', methods=["GET"])
@render_as_json
@authenticate()
def api_v1_settings_scp_public_key():
   data = request.args.to_dict()
   host = validators.get_host('host', data=data)
   port = validators.get_port('port', data=data)
   if host and port:
      return {'keys': util.get_scp_public_key(host, port)}
   else:
      return None

@app.route('/api/v1/settings/scp_fileexport')
@render_as_json
@authenticate()
def api_v1_get_settings_scp_fileexport():
   return util.get_safefile_config()

@app.route('/api/v1/settings/scp_fileexport', methods=["PUT"])
@render_as_json
@authenticate()
def api_v1_settings_scp_fileexport():
   enabled = validators.get_bool('enabled')
   host = validators.get_host('host')
   port = validators.get_port('port')
   user = validators.get_string('user')
   path = validators.get_optional_string('path', required=False) or ''
   password = validators.get_optional_string('password', required=False) or ''
   public_keys = validators.get_string('public_keys')
   util.set_safefile_config(enabled=enabled, host=host, port=port,
                            user=user, path=path, password=password,
                            public_keys=public_keys)
   jobs.restart_safefile_processes.delay()
   return { 'ok': True }

@app.route('/api/v1/decodeCert', methods=['POST'])
@render_as_json
@authenticate()
def api_v1_decode_cert_get_info():
   cert = validators.get_public_cert('cert')
   rv = trust.get_cert_details(cert, True)
   return rv

@app.route('/api/v1/ca_cert', methods=['GET'])
@render_as_json
@authenticate()
def api_v1_ca_cert_get_info():
   cas_path = config.get('general', 'user_installed_cas_path')
   certs = []
   rv = { 'ok': True, 'certs': certs }
   if not os.path.exists(cas_path):
      return rv
   for cert_path in glob.glob(os.path.join(cas_path, '*.crt')):
      try:
         ca_info = trust.get_ca_cert_details(cert_path=cert_path)
         if ca_info.get('certId'):
            ca_info['file_name'] = os.path.basename(cert_path)
            with open(cert_path) as f:
               ca_info['content'] = base64.b64encode(f.read())
            certs.append(ca_info)
      except Exception as ex:
         logging.error('CA cert read error: %s', ex)
   return rv

@app.route('/api/v1/ca_cert/<cert_id>', methods=['POST'])
@render_as_json
@authenticate()
def api_v1_ca_cert_post_upload(cert_id):  # Disregard cert_id given as argument.
   cas_path = config.get('general', 'user_installed_cas_path')
   if not os.path.exists(cas_path):
      try:
         os.makedirs(cas_path)
         logging.info('Created user-installed CAs dir: %s', cas_path)
      except OSError as ex:
         logging.error('Could not create dir user_installed_cas_path: %s (%s)',
                       cas_path, ex)
         return (Error.CA_CERT_MKDIR_FAILED, 400)
   try:
      filename = validators.get_public_ca_cert('cert')
      fingerprint = cert_utils.get_cert_fingerprint_from_file(filename, raise_on_error=True)
   except (validators.InvalidCACertificateError, Exception) as ex:
      logging.error('Invalid CA cert uploaded (error: %s)', ex)
      return (Error.CA_CERT_VALIDATION_FAILED, 400)
   cert_file = cert_utils.fingerprint_to_filename(fingerprint)
   cert_path = os.path.join(cas_path, cert_file)
   if os.path.exists(cert_path):
      logging.error('CA cert path already exists: %s', cert_path)
      return (Error.CA_CERT_ALREADY_EXISTS, 400)
   try:
      os.rename(filename, cert_path)
   except Exception as ex:
      logging.error('Could not rename CA cert file: %s (%s)', cert_path, ex)
      return (Error.CA_CERT_SAVE_FAILED, 400)
   # certificates are not secret and other processes (squid) require read
   # access
   os.chmod(cert_path, os.stat(cert_path).st_mode | stat.S_IRGRP | stat.S_IROTH)
   try:
      cert_utils.sync_ca_certs_to_system(from_dir=cas_path, add_cert=cert_path)
   except Exception as ex:
      logging.error('CA cert sync failed after adding: %s (%s)', cert_path, ex)
      return (Error.CA_CERT_SYNC_FAILED, 400)
   jobs.restart_surrogates.delay()
   return { 'ok': True }

@app.route('/api/v1/ca_cert/<cert_fingerprint>', methods=['DELETE'])
@render_as_json
@authenticate()
def api_v1_ca_cert_delete(cert_fingerprint):
   cas_path = config.get('general', 'user_installed_cas_path')
   if not os.path.exists(cas_path):
      logging.warning('CA cert dir (%s) not found to delete CA cert: %s',
                      cas_path, cert_fingerprint)
      return (Error.CA_CERT_NO_DIR, 400)
   cert_file = cert_utils.fingerprint_to_filename(cert_fingerprint)
   cert_path = os.path.join(cas_path, cert_file)
   if not os.path.exists(cert_path):
      logging.error('Could not find CA cert file to delete: %s', cert_path)
      return (Error.CA_CERT_NOT_FOUND, 400)
   try:
      cert_utils.sync_ca_certs_to_system(from_dir=cas_path, del_cert=cert_path)
   except Exception as ex:
      logging.error('CA cert deletion failed: %s (%s)', cert_path, ex)
      return (Error.CA_CERT_DELETE_FAILED, 400)
   return { 'ok': True }

def _remove_zombie_rq_workers():
   """ remove 'zombie' rq workers from Redis.

    RQ workers do not register death on a hard service stop
    (i.e. stop safeview-all / vcenter shutdown), and persist in redis through
    the next service up as "zombie" workers. This can be the problem when any
    code path iterate through all the RQ active workers.

    Detect 'zombie' workers:
    - RQ stores each worker in the Redis with the key as below:
        <redis_worker_namespace_prefix><subdomain from system's hostname>.<PID>
      i.e
        rq:worker:msip-test-1000-standalone-1-eth0.15059
    - While iterating though all the workers, check can be performed if there is
      any process running with the PID present in the worker key if not then
      simply remove it from the Redis.
    """
   try:
      workers = rq.Worker.all()
      # get PIDs of all running processes at the moment on the system.
      pids = psutil.pids()
      # iterate through all the workers present in the Redis.
      for worker in workers:
         w_pid = int(worker.name.split('.')[-1])
         if worker.get_state() in ['idle', 'busy'] and w_pid not in pids:
            # worker is zombie and should be removed from the Redis.
            # Ref: https://github.com/rq/rq/blob/0.5.5/rq/worker.py#L249
            worker.register_death()
            logging.info('event=remove-zombie-rq-worker worker=%s' %
                         worker.name)
   except Exception as e:
      logging.warning('event=remove-zombie-rq-worker-failed error=%s' % e)


if __name__ == '__main__':
   dashboard_host = config.get('dashboard', 'host')
   dashboard_port = config.getint('dashboard', 'port')
   logging.info("Starting dashboard on %s:%d", dashboard_host, dashboard_port)
   # Make sure user facing default certificates are available
   cert_utils.init()
   # Initialize the database
   db.init()
   # Need to share _globals.api_token_valid_after with dashboard-ng
   db.set_token_valid_after(_globals.api_token_valid_after)
   # watch for configuration changes
   util.start_config_file_watcher()
   # start licensing check
   _start_license_check()
   # remove zombie rq workers
   _remove_zombie_rq_workers()
   # Empty out the failed queue, since we are not doing anything about this
   # anyways. If we don't do this, these will exist in redis forever, even
   # after reset/reboot
   try:
      rq.get_failed_queue().empty()
   except Exception as e:
      logging.warning('Failed to empty queue: %s', e)
   server = HTTPServer(WSGIContainer(app))
   server.listen(dashboard_port, dashboard_host)
   IOLoop.configure(tornado.platform.epoll.EPollIOLoop,
                    time_func=get_monotonic_time)
   IOLoop.instance().start()
